''' opencos.tools.yosys - base class for slang_yosys.py, invio_yosys.py, tabbycad_yosys.py

Contains classes for ToolYosys
'''

# pylint: disable=R0801 # (calling functions with same arguments)

import shutil
import subprocess

from opencos import util
from opencos.eda_base import Tool
from opencos.commands import CommandSynth

class ToolYosys(Tool):
    '''Parent class for ToolTabbyCadYosys, ToolInvioYosys, ToolSlangYosys'''

    _TOOL = 'yosys'
    _EXE = 'yosys'
    _URL = 'https://yosyshq.readthedocs.io/en/latest/'

    yosys_exe = ''
    sta_exe = ''
    sta_version = ''

    def get_versions(self) -> str:
        if self._VERSION:
            return self._VERSION

        path = shutil.which(self._EXE)
        if not path:
            self.error(f'"{self._EXE}" not in path or not installed, see {self._URL}')
        else:
            self.yosys_exe = path

        # Unforunately we don't have a non-PATH friendly support on self._EXE to set
        # where standalone 'sta' is. Even though Yosys has 'sta' internally, Yosys does
        # not fully support timing constraints or .sdc files, so we have to run 'sta'
        # standalone.
        sta_path = shutil.which('sta')
        if sta_path:
            util.debug(f'Also located "sta" via {sta_path}')
            self.sta_exe = sta_path
            sta_version_ret = subprocess.run(
                [self.sta_exe, '-version'], capture_output=True, check=False
            )
            util.debug(f'{self.yosys_exe} {sta_version_ret=}')
            sta_ver = sta_version_ret.stdout.decode('utf-8').split()[0]
            if sta_ver:
                self.sta_version = sta_ver

        version_ret = subprocess.run(
            [self.yosys_exe, '--version'], capture_output=True, check=False
        )
        util.debug(f'{self.yosys_exe} {version_ret=}')

        # Yosys 0.48 (git sha1 aaa534749, clang++ 14.0.0-1ubuntu1.1 -fPIC -O3)
        words = version_ret.stdout.decode('utf-8').split()

        if len(words) < 2:
            self.error(f'{self.yosys_exe} --version: returned unexpected str {version_ret=}')
        self._VERSION = words[1]
        return self._VERSION

    def set_tool_defines(self):
        self.defines.update({
            'OC_TOOL_YOSYS': None
        })
        if 'OC_LIBRARY' not in self.defines:
            self.defines.update({
                'OC_LIBRARY_BEHAVIORAL': None,
                'OC_LIBRARY': "0"
            })


class CommonSynthYosys(CommandSynth, ToolYosys):
    '''Common parent class used by invio_yosys and tabbycad_yosys

    for child classes: CommandSynthInvioYosys and tabbycad_yosys.CommandSynthTabbyCadYosys
    '''

    def __init__(self, config:dict):
        CommandSynth.__init__(self, config=config)
        ToolYosys.__init__(self, config=self.config)

        self.args.update({
            'yosys-synth': 'synth',              # synth_xilinx, synth_altera, etc (see: yosys help)
            'yosys-pre-synth': ['prep', 'proc'], # command run in yosys prior to yosys-synth.
            'yosys-blackbox': [],                # list of modules that yosys will blackbox.
        })

    def do_it(self) -> None:
        self.set_tool_defines()
        self.write_eda_config_and_args()

        if self.is_export_enabled():
            self.do_export()
            return

        self.write_and_run_yosys_f_files()

    def write_and_run_yosys_f_files(self, **kwargs) -> None:
        '''Derived classes must define, to run remainder of do_it() steps'''
        raise NotImplementedError



#!/usr/bin/env python3

'''
Helper pymodule used by eda_deps_bash_completion.bash, extracts valid
targets from DEPS files
'''

import sys
import os
import json

import yaml
import toml

from opencos.deps_helpers import get_deps_markup_file

def get_markup_table_keys(partial_path='./') -> list:
    '''Returns a list of root level keys for DEPS.[yml|yaml|toml|json]

    Does not include DEFAULTS.
    '''
    partial_target = ''
    if not partial_path or partial_path == '.':
        partial_path = './'

    if not os.path.exists(partial_path):
        partial_path, partial_target = os.path.split(partial_path)
    if not partial_path:
        partial_path = './'

    filepath = get_deps_markup_file(base_path=partial_path)
    if not filepath:
        # Couldn't find a DEPS file, let bash completion handle it (with -W words or -G glob)
        return []

    data = {}
    _, file_ext = os.path.splitext(filepath)
    try:
        if file_ext in ['', '.yml', 'yaml']:
            with open(filepath, 'r', encoding='utf-8') as f:
                data = yaml.safe_load(f)
        elif file_ext == '.toml':
            data = toml.load(filepath)
        elif file_ext == '.json':
            with open(filepath, 'r', encoding='utf-8') as f:
                data = json.load(f)
    except Exception: # pylint: disable=broad-exception-caught
        return []

    if not isinstance(data, dict):
        # We found a DEPS file, but it wasn't a table/dict so we can't return root keys
        return []

    # Try to resolve path/to/target/partial_target_
    # -- prepend path information to found targets in path/to/target/DEPS
    prepend = ''
    if partial_path and partial_path != './':
        prepend = partial_path
        if not partial_path.endswith('/'):
            prepend += '/'

    # Return the list of keys w/ prepended path information, and don't include
    # uppercase strings like 'DEFAULTS' or 'METADATA'
    return [
        prepend + x for x in list(data.keys()) if x.startswith(partial_target) and not x.isupper()
    ]


def main() -> None:
    '''Returns None, prints DEPS keys space separated, uses sys.argv for a single
    partial path arg (DEPS file to examine)'''

    if len(sys.argv) > 1:
        partial_path = sys.argv[1]
    else:
        partial_path = './'
    keys = get_markup_table_keys(partial_path)
    print(" ".join(keys))

if __name__ == "__main__":
    main()
